import logging

import hmac
from hashlib import sha256
import requests
from odoo import models, api
from odoo.addons.saas_worker.const import GLOBAL_OAUTH_BASE_URL

_logger = logging.getLogger(__name__)


class ResUsers(models.Model):
    _inherit = "res.users"

    def _session_signature_token_matches(self, reference_token):
        # verify session token matches reference one, also with alternative login for
        # internal users, in order to support transition period where 'trigram'
        # becomes 'trigram@odoo.com'
        # TODO: once we stop supporting dual logins, we can revert this whole commit.

        # retrieve the fields used to generate the session token
        sorted_session_fields = sorted(self._get_session_token_fields())
        session_fields = ", ".join(sorted_session_fields)
        session_fields = f"share, {session_fields}"  # distinguish employees!

        self.env.cr.execute("""SELECT %s, (SELECT value FROM ir_config_parameter WHERE key='database.secret')
                                FROM res_users
                                WHERE id=%%s""" % (session_fields),
                            (self.id,))
        if self.env.cr.rowcount != 1:
            # no match, or too many => try to disable them (either all, or none)
            return False

        data_fields = self.env.cr.fetchone()
        share, *data_fields = data_fields
        # data_fields should have: active, id, login, password, totp_secret, ...
        login = data_fields[2]

        altervative_logins = [login]
        if not share:
            # for employees, support both login forms: tri and tri@odoo.com
            login = login.replace("@odoo.com", "")
            altervative_logins = [login, f"{login}@odoo.com"]

        active, uid, _login, *rest = data_fields
        sid = str(uid)
        for login in altervative_logins:
            # generate hmac key
            key = active, uid, login, *rest
            key = str(key).encode()
            # hmac the session id (which is UID)
            data = sid.encode()
            h = hmac.new(key, data, sha256)

            if h.hexdigest() == reference_token:
                # match => no need to disable this user
                return True

        # default is to disable the login, in case of failure, we'd rather disable too many users
        return False


    def _sync_sess_change(self, session_tokens):
        """
        Check the session token of the uids. Loop on the list of tuples (uid, session_token) and
        reset the password if the session_token is different.

        :param list session_tokens: the list of (uid, session_token) to verify
        """

        _logger.debug("Sync session change for uids %s", [s[0] for s in session_tokens])
        if not session_tokens:
            return

        # Build the list of changed tokens
        Users = self.env["res.users"].sudo()
        uids_to_disable = [
            uid
            for (uid, session_token) in session_tokens
            if not Users.browse(uid)._session_signature_token_matches(session_token)
        ]
        if not uids_to_disable:
            return

        # Reset access for users with changed session token
        self.env.cr.execute(
            """
                UPDATE res_users
                SET password = NULL
                WHERE id IN %s RETURNING id
        """,
            (tuple(uids_to_disable),),
        )
        reset_uids = [r[0] for r in self.env.cr.fetchall()]
        if reset_uids:
            _logger.info("Reset access for users with ids %s", reset_uids)
        else:
            _logger.debug("No access reset")

    def _cron_sync_outdated(self):
        """
        Synchronize users who changed their password recently.
        Users who changed their passwords after the database copy have their password
        erased (therefore they lose their access).
        """

        # Retrieve outdated sessions on accounts.odoo.com. Since the only source of truth is
        # the production database, we use GLOBAL_OAUTH_BASE_URL which is not overrided for test
        # duplicates.
        # Do not sign the request since this will run in the TEST environment. From the client
        # perspective, the name is '<server_name>.test.odoo.com', while from accounts the name
        # is '<server_name>.odoo.com'. This will cause the signature to be invalid.
        res = requests.post(GLOBAL_OAUTH_BASE_URL + "/_internal/outdated_users", timeout=10)
        res.raise_for_status()
        res_json = res.json()

        # Reset passwords.
        self._sync_sess_change(res_json.get("sess_change"))

    @classmethod
    def authenticate(cls, db, login, password, user_agent_env):
        uid = super().authenticate(db, login, password, user_agent_env=user_agent_env)

        if user_agent_env and user_agent_env.get('base_location'):
            with cls.pool.cursor() as cr:
                env = api.Environment(cr, uid, {})
                if env.user.has_group('base.group_system'):  # Successfully logged in as system user!
                    domain = user_agent_env['base_location']
                    base_url = env['ir.config_parameter'].get_param('web.base.url', '')
                    if domain.startswith('https://www.') and base_url != domain:
                        env['openerp.enterprise.tool']._set_domain(domain)
                        _logger.info("web.base.url updated to %s", domain)
        return uid
